@@ -15,12 +15,14 @@
 # You should have received a copy of the GNU General Public License
 # along with OCRmyPDF.  If not, see <http://www.gnu.org/licenses/>.
 
-import os
+import logging
 from contextlib import suppress
 from pathlib import Path
+from typing import Optional
 
 import pikepdf
 
+log = logging.getLogger(__name__)
 MAX_REPLACE_PAGES = 100
 
 
@@ -88,99 +90,10 @@ def strip_invisible_text(pdf, page):
     page.Contents = pikepdf.Stream(pdf, content_stream)
 
 
-def _graft_text_layer(
-    *, pdf_base, page_num, text, font, font_key, procset, rotation, strip_old_text, log
-):
-    """Insert the text layer from text page 0 on to pdf_base at page_num"""
-
-    log.debug("Grafting")
-    if Path(text).stat().st_size == 0:
-        return
-
-    # This is a pointer indicating a specific page in the base file
-    pdf_text = pikepdf.open(text)
-    pdf_text_contents = pdf_text.pages[0].Contents.read_bytes()
-
-    base_page = pdf_base.pages.p(page_num)
-
-    # The text page always will be oriented up by this stage but the original
-    # content may have a rotation applied. Wrap the text stream with a rotation
-    # so it will be oriented the same way as the rest of the page content.
-    # (Previous versions OCRmyPDF rotated the content layer to match the text.)
-    mediabox = [float(pdf_text.pages[0].MediaBox[v]) for v in range(4)]
-    wt, ht = mediabox[2] - mediabox[0], mediabox[3] - mediabox[1]
-
-    mediabox = [float(base_page.MediaBox[v]) for v in range(4)]
-    wp, hp = mediabox[2] - mediabox[0], mediabox[3] - mediabox[1]
-
-    translate = pikepdf.PdfMatrix().translated(-wt / 2, -ht / 2)
-    untranslate = pikepdf.PdfMatrix().translated(wp / 2, hp / 2)
-    corner = pikepdf.PdfMatrix().translated(mediabox[0], mediabox[1])
-    # -rotation because the input is a clockwise angle and this formula
-    # uses CCW
-    rotation = -rotation % 360
-    rotate = pikepdf.PdfMatrix().rotated(rotation)
-
-    # Because of rounding of DPI, we might get a text layer that is not
-    # identically sized to the target page. Scale to adjust. Normally this
-    # is within 0.998.
-    if rotation in (90, 270):
-        wt, ht = ht, wt
-    scale_x = wp / wt
-    scale_y = hp / ht
-
-    # log.debug('%r', scale_x, scale_y)
-    scale = pikepdf.PdfMatrix().scaled(scale_x, scale_y)
-
-    # Translate the text so it is centered at (0, 0), rotate it there, adjust
-    # for a size different between initial and text PDF, then untranslate, and
-    # finally move the lower left corner to match the mediabox
-    ctm = translate @ rotate @ scale @ untranslate @ corner
-
-    pdf_text_contents = b'q %s cm\n' % ctm.encode() + pdf_text_contents + b'\nQ\n'
-
-    new_text_layer = pikepdf.Stream(pdf_base, pdf_text_contents)
-
-    if strip_old_text:
-        strip_invisible_text(pdf_base, base_page)
-
-    base_page.page_contents_add(new_text_layer, prepend=True)
-
-    _update_page_resources(
-        page=base_page, font=font, font_key=font_key, procset=procset
-    )
-    pdf_text.close()
-
-
-def _find_font(text, pdf_base):
-    """Copy a font from the filename text into pdf_base"""
-
-    font, font_key = None, None
-    possible_font_names = ('/f-0-0', '/F1')
-    try:
-        with pikepdf.open(text) as pdf_text:
-            try:
-                pdf_text_fonts = pdf_text.pages[0].Resources.get('/Font', {})
-            except (AttributeError, IndexError, KeyError):
-                return None, None
-            for f in possible_font_names:
-                pdf_text_font = pdf_text_fonts.get(f, None)
-                if pdf_text_font is not None:
-                    font_key = f
-                    break
-            if pdf_text_font:
-                font = pdf_base.copy_foreign(pdf_text_font)
-            return font, font_key
-    except (FileNotFoundError, pikepdf.PdfError):
-        # PdfError occurs if a 0-length file is written e.g. due to OCR timeout
-        return None, None
-
-
 class OcrGrafter:
     def __init__(self, context):
         self.context = context
-        self.log = context.log
-        self.path_base = Path(context.origin).resolve()
+        self.path_base = context.origin
 
         self.pdf_base = pikepdf.open(self.path_base)
         self.font, self.font_key = None, None
@@ -195,10 +108,16 @@ class OcrGrafter:
         self.emplacements = 1
         self.interim_count = 0
 
-    def graft_page(self, page_result):
-        pageno, image, text, _sidecar, autorotate_correction = page_result
-        if text and not self.font:
-            self.font, self.font_key = _find_font(text, self.pdf_base)
+    def graft_page(
+        self,
+        *,
+        pageno: int,
+        image: Optional[Path],
+        textpdf: Optional[Path],
+        autorotate_correction: int,
+    ):
+        if textpdf and not self.font:
+            self.font, self.font_key = self._find_font(textpdf)
 
         emplaced_page = False
         content_rotation = self.pdfinfo[pageno].rotation
@@ -206,7 +125,7 @@ class OcrGrafter:
         if path_image is not None and path_image != self.path_base:
             # We are updating the old page with a rasterized PDF of the new
             # page (without changing objgen, to preserve references)
-            self.log.debug("Emplacement update")
+            log.debug("Emplacement update")
             with pikepdf.open(image) as pdf_image:
                 self.emplacements += 1
                 foreign_image_page = pdf_image.pages[0]
@@ -220,25 +139,23 @@ class OcrGrafter:
             content_rotation = autorotate_correction
         text_rotation = autorotate_correction
         text_misaligned = (text_rotation - content_rotation) % 360
-        self.log.debug(
+        log.debug(
             f"Rotations for page {pageno}: [text, auto, misalign, content] = "
             f"{text_rotation}, {autorotate_correction}, "
             f"{text_misaligned}, {content_rotation}"
         )
 
-        if text and self.font:
+        if textpdf and self.font:
             # Graft the text layer onto this page, whether new or old
             strip_old = self.context.options.redo_ocr
-            _graft_text_layer(
-                pdf_base=self.pdf_base,
+            self._graft_text_layer(
                 page_num=pageno + 1,
-                text=text,
+                textpdf=textpdf,
                 font=self.font,
                 font_key=self.font_key,
                 rotation=text_misaligned,
                 procset=self.procset,
                 strip_old_text=strip_old,
-                log=self.log,
             )
 
         # Correct the rotation if applicable
@@ -250,10 +167,13 @@ class OcrGrafter:
             self.save_and_reload()
 
     def save_and_reload(self):
-        # Periodically save and reload the Pdf object. This will keep a
-        # lid on our memory usage for very large files. Attach the font to
-        # page 1 even if page 1 doesn't use it, so we have a way to get it
-        # back.
+        """Save and reload the Pdf.
+
+        This will keep a lid on our memory usage for very large files. Attach
+        the font to page 1 even if page 1 doesn't use it, so we have a way to get it
+        back.
+        """
+
         page0 = self.pdf_base.pages[0]
         _update_page_resources(
             page=page0, font=self.font, font_key=self.font_key, procset=self.procset
@@ -264,12 +184,14 @@ class OcrGrafter:
         # {interim_count} is the opened file we were updateing
         # {interim_count - 1} can be deleted
         # {interim_count + 1} is the new file will produce and open
-        old_file = self.output_file + f'_working{self.interim_count - 1}.pdf'
+        old_file = self.output_file.with_suffix(f'.working{self.interim_count - 1}.pdf')
         if not self.context.options.keep_temporary_files:
             with suppress(FileNotFoundError):
-                os.unlink(old_file)
+                old_file.unlink()
 
-        next_file = self.output_file + f'_working{self.interim_count + 1}.pdf'
+        next_file = self.output_file.with_suffix(
+            f'.working{self.interim_count + 1}.pdf'
+        )
         self.pdf_base.save(next_file)
         self.pdf_base.close()
 
@@ -282,3 +204,98 @@ class OcrGrafter:
         self.pdf_base.save(self.output_file)
         self.pdf_base.close()
         return self.output_file
+
+    def _find_font(self, text):
+        """Copy a font from the filename text into pdf_base"""
+
+        font, font_key = None, None
+        possible_font_names = ('/f-0-0', '/F1')
+        try:
+            with pikepdf.open(text) as pdf_text:
+                try:
+                    pdf_text_fonts = pdf_text.pages[0].Resources.get('/Font', {})
+                except (AttributeError, IndexError, KeyError):
+                    return None, None
+                for f in possible_font_names:
+                    pdf_text_font = pdf_text_fonts.get(f, None)
+                    if pdf_text_font is not None:
+                        font_key = f
+                        break
+                if pdf_text_font:
+                    font = self.pdf_base.copy_foreign(pdf_text_font)
+                return font, font_key
+        except (FileNotFoundError, pikepdf.PdfError):
+            # PdfError occurs if a 0-length file is written e.g. due to OCR timeout
+            return None, None
+
+    def _graft_text_layer(
+        self,
+        *,
+        page_num: int,
+        textpdf: Path,
+        font: pikepdf.Object,
+        font_key: pikepdf.Object,
+        procset: pikepdf.Object,
+        rotation: int,
+        strip_old_text: bool,
+    ):
+        """Insert the text layer from text page 0 on to pdf_base at page_num"""
+
+        log.debug("Grafting")
+        if Path(textpdf).stat().st_size == 0:
+            return
+
+        # This is a pointer indicating a specific page in the base file
+        with pikepdf.open(textpdf) as pdf_text:
+            pdf_text_contents = pdf_text.pages[0].Contents.read_bytes()
+
+            base_page = self.pdf_base.pages.p(page_num)
+
+            # The text page always will be oriented up by this stage but the original
+            # content may have a rotation applied. Wrap the text stream with a rotation
+            # so it will be oriented the same way as the rest of the page content.
+            # (Previous versions OCRmyPDF rotated the content layer to match the text.)
+            mediabox = [float(pdf_text.pages[0].MediaBox[v]) for v in range(4)]
+            wt, ht = mediabox[2] - mediabox[0], mediabox[3] - mediabox[1]
+
+            mediabox = [float(base_page.MediaBox[v]) for v in range(4)]
+            wp, hp = mediabox[2] - mediabox[0], mediabox[3] - mediabox[1]
+
+            translate = pikepdf.PdfMatrix().translated(-wt / 2, -ht / 2)
+            untranslate = pikepdf.PdfMatrix().translated(wp / 2, hp / 2)
+            corner = pikepdf.PdfMatrix().translated(mediabox[0], mediabox[1])
+            # -rotation because the input is a clockwise angle and this formula
+            # uses CCW
+            rotation = -rotation % 360
+            rotate = pikepdf.PdfMatrix().rotated(rotation)
+
+            # Because of rounding of DPI, we might get a text layer that is not
+            # identically sized to the target page. Scale to adjust. Normally this
+            # is within 0.998.
+            if rotation in (90, 270):
+                wt, ht = ht, wt
+            scale_x = wp / wt
+            scale_y = hp / ht
+
+            # log.debug('%r', scale_x, scale_y)
+            scale = pikepdf.PdfMatrix().scaled(scale_x, scale_y)
+
+            # Translate the text so it is centered at (0, 0), rotate it there, adjust
+            # for a size different between initial and text PDF, then untranslate, and
+            # finally move the lower left corner to match the mediabox
+            ctm = translate @ rotate @ scale @ untranslate @ corner
+
+            pdf_text_contents = (
+                b'q %s cm\n' % ctm.encode() + pdf_text_contents + b'\nQ\n'
+            )
+
+            new_text_layer = pikepdf.Stream(self.pdf_base, pdf_text_contents)
+
+            if strip_old_text:
+                strip_invisible_text(self.pdf_base, base_page)
+
+            base_page.page_contents_add(new_text_layer, prepend=True)
+
+            _update_page_resources(
+                page=base_page, font=font, font_key=font_key, procset=procset
+            )
