@@ -18,44 +18,24 @@
 import logging
 import os
 import sys
-from contextlib import suppress
 from enum import IntEnum
 from pathlib import Path
-from typing import Dict, List
+from typing import BinaryIO, Iterable, Union
 
-from tqdm import tqdm
+from ocrmypdf._logging import PageNumberFilter, TqdmConsole
+from ocrmypdf._plugin_manager import get_plugin_manager
+from ocrmypdf._sync import run_pipeline
+from ocrmypdf._validation import check_options
+from ocrmypdf.cli import ArgumentParser, get_parser
+from ocrmypdf.helpers import is_iterable_notstr
 
-from ._sync import run_pipeline
-from ._validation import check_options
-from .cli import parser
+try:
+    import coloredlogs
+except ModuleNotFoundError:
+    coloredlogs = None
 
 
-class TqdmConsole:
-    """Wrapper to log messages in a way that is compatible with tqdm progress bar
-
-    This routes log messages through tqdm so that it can print them above the
-    progress bar, and then refresh the progress bar, rather than overwriting
-    it which looks messy.
-
-    For some reason Python 3.6 prints extra empty messages from time to time,
-    so we suppress those.
-    """
-
-    def __init__(self, file):
-        self.file = file
-        self.py36 = sys.version_info[0:2] == (3, 6)
-
-    def write(self, msg):
-        # When no progress bar is active, tqdm.write() routes to print()
-        if self.py36:
-            if msg.strip() != '':
-                tqdm.write(msg.rstrip(), end='\n', file=self.file)
-        else:
-            tqdm.write(msg.rstrip(), end='\n', file=self.file)
-
-    def flush(self):
-        with suppress(AttributeError):
-            self.file.flush()
+PathOrIO = Union[BinaryIO, os.PathLike, str, bytes]
 
 
 class Verbosity(IntEnum):
@@ -67,7 +47,11 @@ class Verbosity(IntEnum):
     debug_all = 2  #: More detailed debugging from ocrmypdf and dependent modules
 
 
-def configure_logging(verbosity, progress_bar_friendly=True, manage_root_logger=False):
+def configure_logging(
+    verbosity: Verbosity,
+    progress_bar_friendly: bool = True,
+    manage_root_logger: bool = False,
+):
     """Set up logging.
 
     Library users may wish to use this function if they want their log output to be
@@ -94,6 +78,7 @@ def configure_logging(verbosity, progress_bar_friendly=True, manage_root_logger=
     """
 
     prefix = '' if manage_root_logger else 'ocrmypdf'
+
     log = logging.getLogger(prefix)
     log.setLevel(logging.DEBUG)
 
@@ -109,9 +94,25 @@ def configure_logging(verbosity, progress_bar_friendly=True, manage_root_logger=
     else:
         console.setLevel(logging.INFO)
 
-    formatter = logging.Formatter('%(levelname)7s - %(message)s')
+    console.addFilter(PageNumberFilter())
+
     if verbosity >= 2:
-        formatter = logging.Formatter('%(name)s - %(levelname)7s - %(message)s')
+        fmt = '%(levelname)7s %(name)s -%(pageno)s %(message)s'
+    else:
+        fmt = '%(pageno)s%(message)s'
+
+    use_colors = progress_bar_friendly
+    if not coloredlogs:
+        use_colors = False
+    if use_colors:
+        if os.name == 'nt':
+            use_colors = coloredlogs.enable_ansi_support()
+        if use_colors:
+            use_colors = coloredlogs.terminal_supports_colors()
+    if use_colors:
+        formatter = coloredlogs.ColoredFormatter(fmt=fmt)
+    else:
+        formatter = logging.Formatter(fmt=fmt)
 
     console.setFormatter(formatter)
     log.addHandler(console)
@@ -128,7 +129,9 @@ def configure_logging(verbosity, progress_bar_friendly=True, manage_root_logger=
     return log
 
 
-def create_options(*, input_file, output_file, **kwargs):
+def create_options(
+    *, input_file: PathOrIO, output_file: PathOrIO, parser: ArgumentParser, **kwargs
+):
     cmdline = []
     deferred = []
 
@@ -138,7 +141,7 @@ def create_options(*, input_file, output_file, **kwargs):
 
         # These arguments with special handling for which we bypass
         # argparse
-        if arg in {'tesseract_env', 'progress_bar'}:
+        if arg in {'progress_bar', 'plugins'}:
             deferred.append((arg, val))
             continue
 
@@ -150,6 +153,12 @@ def create_options(*, input_file, output_file, **kwargs):
                 cmdline.append(f"--{cmd_style_arg}")
             continue
 
+        if is_iterable_notstr(val):
+            for elem in val:
+                cmdline.append(f"--{cmd_style_arg}")
+                cmdline.append(elem)
+            continue
+
         # We have a parameter
         cmdline.append(f"--{cmd_style_arg}")
         if isinstance(val, (int, float)):
@@ -161,26 +170,33 @@ def create_options(*, input_file, output_file, **kwargs):
         else:
             raise TypeError(f"{arg}: {val} ({type(val)})")
 
-    cmdline.append(str(input_file))
-    cmdline.append(str(output_file))
+    try:
+        cmdline.append(os.fspath(input_file))
+    except TypeError:
+        cmdline.append('stream://input_file')
+    try:
+        cmdline.append(os.fspath(output_file))
+    except TypeError:
+        cmdline.append('stream://output_file')
 
-    parser.api_mode = True
+    parser._api_mode = True
     options = parser.parse_args(cmdline)
     for keyword, val in deferred:
         setattr(options, keyword, val)
 
-    # If we are running a Tesseract spoof, ensure it knows what the input file is
-    if os.environ.get('PYTEST_CURRENT_TEST') and options.tesseract_env:
-        options.tesseract_env['_OCRMYPDF_TEST_INFILE'] = os.fspath(input_file)
+    if options.input_file == 'stream://input_file':
+        options.input_file = input_file
+    if options.output_file == 'stream://output_file':
+        options.output_file = output_file
 
     return options
 
 
 def ocr(  # pylint: disable=unused-argument
-    input_file: os.PathLike,
-    output_file: os.PathLike,
+    input_file: PathOrIO,
+    output_file: PathOrIO,
     *,
-    language: List[str] = None,
+    language: Iterable[str] = None,
     image_dpi: int = None,
     output_type=None,
     sidecar: os.PathLike = None,
@@ -210,7 +226,7 @@ def ocr(  # pylint: disable=unused-argument
     jbig2_page_group_size: int = None,
     pages: str = None,
     max_image_mpixels: float = None,
-    tesseract_config: List[str] = None,
+    tesseract_config: Iterable[str] = None,
     tesseract_pagesegmode: int = None,
     tesseract_oem: int = None,
     pdf_renderer=None,
@@ -220,9 +236,10 @@ def ocr(  # pylint: disable=unused-argument
     user_words: os.PathLike = None,
     user_patterns: os.PathLike = None,
     fast_web_view: float = None,
+    plugins: Iterable[str] = None,
     keep_temporary_files: bool = None,
     progress_bar: bool = None,
-    tesseract_env: Dict[str, str] = None,
+    **kwargs,
 ):
     """Run OCRmyPDF on one PDF or image.
 
@@ -230,10 +247,24 @@ def ocr(  # pylint: disable=unused-argument
     A few specific arguments are discussed here:
 
     Args:
-        use_threads (bool): Use worker threads instead of processes. This reduces
+        use_threads: Use worker threads instead of processes. This reduces
             performance but may make debugging easier since it is easier to set
             breakpoints.
-        tesseract_env (dict): Override environment variables for Tesseract
+        input_file: If a :class:`pathlib.Path`, ``str`` or ``bytes``, this is
+            interpreted as file system path to the input file. If the object
+            appears to be a readable stream (with methods such as ``.read()``
+            and ``.seek()``), the object will be read in its entirety and saved to
+            a temporary file. If ``input_file`` is  ``"-"``, standard input will be
+            read.
+        output_file: If a :class:`pathlib.Path`, ``str`` or ``bytes``, this is
+            interpreted as file system path to the output file. If the object
+            appears to be a writable stream (with methods such as ``.read()`` and
+            ``.seek()``), the output will be written to this stream. If
+            ``output_file`` is ``"-"``, the output will be written to ``sys.stdout``
+            (provided that standard output does not seem to be a terminal device).
+            When a stream is used as output, whether via a writable object or
+            ``"-"``, some final validation steps are not performed (we do not read
+            back the stream after it is written).
     Raises:
         ocrmypdf.PdfMergeFailedError: If the input PDF is malformed, preventing merging
             with the OCR layer.
@@ -257,7 +288,20 @@ def ocr(  # pylint: disable=unused-argument
     Returns:
         :class:`ocrmypdf.ExitCode`
     """
+    if not plugins:
+        plugins = []
+    else:
+        plugins = list(plugins)
 
-    options = create_options(**locals())
-    check_options(options)
-    return run_pipeline(options, api=True)
+    parser = get_parser()
+    _plugin_manager = get_plugin_manager(plugins)
+    _plugin_manager.hook.add_options(parser=parser)  # pylint: disable=no-member
+
+    create_options_kwargs = {
+        k: v for k, v in locals().items() if not k.startswith('_') and k != 'kwargs'
+    }
+    create_options_kwargs.update(kwargs)
+
+    options = create_options(**create_options_kwargs)
+    check_options(options, _plugin_manager)
+    return run_pipeline(options=options, plugin_manager=_plugin_manager, api=True)
