@@ -20,23 +20,58 @@ import multiprocessing
 import os
 import shutil
 import warnings
+from collections import namedtuple
 from collections.abc import Iterable
 from contextlib import suppress
 from functools import wraps
+from io import StringIO
+from math import isclose
 from pathlib import Path
+from typing import Any, Sequence, TypeVar
+
+import pikepdf
 
 log = logging.getLogger(__name__)
 
 
-def safe_symlink(input_file, soft_link_name, *args, **kwargs):
+class Resolution(namedtuple('Resolution', ('x', 'y'))):
+    """The number of pixels per inch in each 2D direction."""
+
+    __slots__ = ()
+
+    def round(self, ndigits: int):
+        return Resolution(round(self.x, ndigits), round(self.y, ndigits))
+
+    def to_int(self):
+        return Resolution(int(round(self.x)), int(round(self.y)))
+
+    @property
+    def is_square(self) -> bool:
+        return isclose(self.x, self.y, rel_tol=1e-3)
+
+    def take_max(self, vals, yvals=None):
+        if yvals is not None:
+            return Resolution(max(self.x, *vals), max(self.y, *yvals))
+        max_x, max_y = self.x, self.y
+        for x, y in vals:
+            max_x = max(x, max_x)
+            max_y = max(y, max_y)
+        return Resolution(max_x, max_y)
+
+    def flip_axis(self):
+        return Resolution(self.y, self.x)
+
+    def __str__(self):
+        return f"{self.x:f}x{self.y:f}"
+
+    def __repr__(self):
+        return f"Resolution({self.x}x{self.y} dpi)"
+
+
+def safe_symlink(input_file: os.PathLike, soft_link_name: os.PathLike):
     """
     Helper function: relinks soft symbolic link if necessary
     """
-    if len(args) == 1 and isinstance(args[0], logging.Logger):
-        log.warning("Deprecated: safe_symlink(,log)")
-    if 'log' in kwargs:
-        log.warning('Deprecated: safe_symlink(...log=)')
-
     input_file = os.fspath(input_file)
     soft_link_name = os.fspath(soft_link_name)
 
@@ -72,21 +107,30 @@ def safe_symlink(input_file, soft_link_name, *args, **kwargs):
     os.symlink(os.path.abspath(input_file), soft_link_name)
 
 
-def is_iterable_notstr(thing):
+def samefile(f1: os.PathLike, f2: os.PathLike):
+    if os.name == 'nt':
+        return f1 == f2
+    else:
+        return os.path.samefile(f1, f2)
+
+
+def is_iterable_notstr(thing: Any) -> bool:
+    """Is this is an iterable type, other than a string?"""
     return isinstance(thing, Iterable) and not isinstance(thing, str)
 
 
-def monotonic(L):
+def monotonic(L: Sequence) -> bool:
     """Does list increase monotonically?"""
     return all(b > a for a, b in zip(L, L[1:]))
 
 
-def page_number(input_file):
+def page_number(input_file: os.PathLike) -> int:
     """Get one-based page number implied by filename (000002.pdf -> 2)"""
     return int(os.path.basename(os.fspath(input_file))[0:6])
 
 
-def available_cpu_count():
+def available_cpu_count() -> int:
+    """Returns number of CPUs in the system."""
     try:
         return multiprocessing.cpu_count()
     except NotImplementedError:
@@ -97,7 +141,7 @@ def available_cpu_count():
     return 1
 
 
-def is_file_writable(test_file):
+def is_file_writable(test_file: os.PathLike) -> bool:
     """Intentionally racy test if target is writable.
 
     We intend to write to the output file if and only if we succeed and
@@ -105,11 +149,7 @@ def is_file_writable(test_file):
     the location is writable.
     """
     try:
-        if not isinstance(test_file, Path):
-            p = Path(test_file)
-        else:
-            p = test_file
-
+        p = Path(test_file)
         if p.is_symlink():
             p = p.resolve(strict=False)
 
@@ -136,6 +176,60 @@ def is_file_writable(test_file):
         return False
 
 
+def check_pdf(input_file: Path) -> bool:
+    """Check if a PDF complies with the PDF specification.
+
+    Checks for proper formatting and proper linearization.
+    """
+    pdf = None
+    try:
+        pdf = pikepdf.open(input_file)
+    except pikepdf.PdfError as e:
+        log.error(e)
+        return False
+    else:
+        messages = pdf.check()
+        for msg in messages:
+            if 'error' in msg.lower():
+                log.error(msg)
+            else:
+                log.warning(msg)
+
+        sio = StringIO()
+        linearize = None
+        try:
+            pdf.check_linearization(sio)
+        except RuntimeError:
+            pass
+        else:
+            linearize = sio.getvalue()
+            if linearize:
+                log.warning(linearize)
+
+        if not messages and not linearize:
+            return True
+        return False
+    finally:
+        if pdf:
+            pdf.close()
+
+
+T = TypeVar('T')
+
+
+def clamp(n: T, smallest: T, largest: T) -> T:
+    """Clamps the value of n to between smallest and largest."""
+    return max(smallest, min(n, largest))
+
+
+def pikepdf_enable_mmap():
+    try:
+        if pikepdf._qpdf.set_access_default_mmap(True):
+            log.debug("pikepdf mmap enabled")
+    except AttributeError:
+        log.debug("pikepdf mmap not available")
+
+
 def deprecated(func):
     """Warn that function is deprecated"""
 
