@@ -24,18 +24,20 @@ import argparse
 import logging
 import os
 import sys
+import threading
 import warnings
+from collections import deque
 from collections.abc import Sequence
 from contextlib import suppress
 from ctypes.util import find_library
 from functools import lru_cache
-from io import BytesIO
+from io import BytesIO, UnsupportedOperation
 from os import fspath
 from tempfile import TemporaryFile
 
-from .exceptions import MissingDependencyError
-from .exec import shim_paths_with_program_files
-from .lib._leptonica import ffi
+from ocrmypdf.exceptions import MissingDependencyError
+from ocrmypdf.lib._leptonica import ffi
+from ocrmypdf.subprocess import shim_paths_with_program_files
 
 # pylint: disable=protected-access
 
@@ -75,7 +77,7 @@ except ffi.error as e:
     ) from e
 
 
-class _LeptonicaErrorTrap:
+class _LeptonicaErrorTrap_Redirect:
     """
     Context manager to trap errors reported by Leptonica.
 
@@ -90,20 +92,21 @@ class _LeptonicaErrorTrap:
 
     """
 
+    leptonica_lock = threading.Lock()
+
     def __init__(self):
         self.tmpfile = None
         self.copy_of_stderr = -1
         self.no_stderr = False
 
     def __enter__(self):
-        from io import UnsupportedOperation
-
         self.tmpfile = TemporaryFile()
 
         # Save the old stderr, and redirect stderr to temporary file
-        with suppress(AttributeError):
-            sys.stderr.flush()
+        self.leptonica_lock.acquire()
         try:
+            with suppress(AttributeError):
+                sys.stderr.flush()
             self.copy_of_stderr = os.dup(sys.stderr.fileno())
             os.dup2(self.tmpfile.fileno(), sys.stderr.fileno(), inheritable=False)
         except AttributeError:
@@ -115,7 +118,10 @@ class _LeptonicaErrorTrap:
             os.dup2(self.tmpfile.fileno(), 2, inheritable=False)
         except UnsupportedOperation:
             self.copy_of_stderr = None
-        return
+        except Exception:
+            self.leptonica_lock.release()
+            raise
+        return self
 
     def __exit__(self, exc_type, exc_value, traceback):
         # Restore old stderr
@@ -132,6 +138,8 @@ class _LeptonicaErrorTrap:
         self.tmpfile.seek(0)  # Cursor will be at end, so move back to beginning
         leptonica_output = self.tmpfile.read().decode(errors='replace')
         self.tmpfile.close()
+        self.leptonica_lock.release()
+
         # If there are Python errors, record them
         if exc_type:
             logger.warning(leptonica_output)
@@ -149,6 +157,62 @@ class _LeptonicaErrorTrap:
         return False
 
 
+tls = threading.local()
+tls.trap = None
+
+
+@ffi.callback("void(char *)")
+def _stderr_handler(cstr):
+    msg = ffi.string(cstr).decode(errors='replace')
+    if msg.startswith("Error"):
+        logger.error(msg)
+    elif msg.startswith("Warning"):
+        logger.warning(msg)
+    else:
+        logger.debug(msg)
+    if tls.trap is not None:
+        tls.trap.append(msg)
+    return
+
+
+class _LeptonicaErrorTrap_Queue:
+    def __init__(self):
+        self.queue = deque()
+
+    def __enter__(self):
+        self.queue.clear()
+        tls.trap = self.queue
+
+    def __exit__(self, exc_type, exc_value, traceback):
+        tls.trap = None
+        output = ''.join(self.queue)
+        self.queue.clear()
+
+        # If there are Python errors, record them
+        if exc_type:
+            logger.warning(output)
+
+        if 'Error' in output:
+            if 'image file not found' in output:
+                raise FileNotFoundError()
+            if 'pixWrite: stream not opened' in output:
+                raise LeptonicaIOError()
+            if 'index not valid' in output:
+                raise IndexError()
+            raise LeptonicaError(output)
+        return False
+
+
+try:
+    lept.leptSetStderrHandler(_stderr_handler)
+except ffi.error:
+    # Pre-1.79 Leptonica does not have leptSetStderrHandler
+    _LeptonicaErrorTrap = _LeptonicaErrorTrap_Redirect
+else:
+    # 1.79 have this new symbol
+    _LeptonicaErrorTrap = _LeptonicaErrorTrap_Queue
+
+
 class LeptonicaError(Exception):
     pass
 
@@ -351,7 +415,7 @@ class Pix(LeptonicaObject):
             py_file.write(buffer)
 
     @classmethod
-    def frompil(self, pillow_image):
+    def frompil(cls, pillow_image):
         """Create a copy of a PIL.Image from this Pix"""
         bio = BytesIO()
         pillow_image.save(bio, format='png', compress_level=1)
@@ -363,7 +427,7 @@ class Pix(LeptonicaObject):
 
     def topil(self):
         """Returns a PIL.Image version of this Pix"""
-        from PIL import Image
+        from PIL import Image  # pylint: disable=import-outside-toplevel
 
         # Leptonica manages data in words, so it implicitly does an endian
         # swap.  Tell Pillow about this when it reads the data.
@@ -534,16 +598,7 @@ class Pix(LeptonicaObject):
             )
             return Pix(thresh_pix)
 
-    def crop_to_foreground(
-        self,
-        threshold=128,
-        mindist=70,
-        erasedist=30,
-        pagenum=0,
-        showmorph=0,
-        display=0,
-        pdfdir=ffi.NULL,
-    ):
+    def crop_to_foreground(self, threshold=128, mindist=70, erasedist=30, showmorph=0):
         if get_leptonica_version() < 'leptonica-1.76':
             # Leptonica 1.76 changed the API for pixFindPageForeground; we don't
             # support the old version
@@ -866,6 +921,8 @@ def get_leptonica_version():
     Caveat: Leptonica expects the caller to free this memory.  We don't,
     since that would involve binding to libc to access libc.free(),
     a pointless effort to reclaim 100 bytes of memory.
+
+    Reminder that this returns "leptonica-1.xx" or "leptonica-1.yy.0".
     """
     return ffi.string(lept.getLeptonicaVersion()).decode()
 
