@@ -443,6 +443,12 @@ pixReadBarcodes(PIXA     *pixa,
                 SARRAY  **psaw,
                 l_int32 debugflag);
 
+PIX *
+pixGenHalftoneMask(PIX      *pixs,
+                        PIX     **ppixtext,
+                        l_int32  *phtfound,
+                        PIXA     *pixadb);
+
 l_int32
 l_generateCIDataForPdf(const char *fname,
                        PIX *pix,
@@ -503,6 +509,8 @@ void selDestroy ( SEL **psel );
 l_int32
 setMsgSeverity(l_int32 newsev);
 
+void
+leptSetStderrHandler(void (*handler)(const char *));
 """
 )
 
