@@ -15,52 +15,80 @@
 # You should have received a copy of the GNU General Public License
 # along with OCRmyPDF.  If not, see <http://www.gnu.org/licenses/>.
 
-import concurrent.futures
+import logging
 import sys
 import tempfile
 from collections import defaultdict
+from functools import partial
 from os import fspath
 from pathlib import Path
+from typing import (
+    Any,
+    Callable,
+    Dict,
+    Iterator,
+    List,
+    MutableSet,
+    NamedTuple,
+    NewType,
+    Optional,
+    Sequence,
+    Tuple,
+    Union,
+)
 
 import pikepdf
-from pikepdf import Dictionary, Name
+from pikepdf import Dictionary, Name, Object, Pdf, PdfImage
 from PIL import Image
 from tqdm import tqdm
 
-from . import leptonica
-from ._jobcontext import PDFContext
-from .exceptions import OutputFileAccessError
-from .exec import jbig2enc, pngquant
-from .helpers import safe_symlink
+from ocrmypdf import leptonica
+from ocrmypdf._concurrent import exec_progress_pool
+from ocrmypdf._exec import jbig2enc, pngquant
+from ocrmypdf._jobcontext import PdfContext
+from ocrmypdf.exceptions import OutputFileAccessError
+from ocrmypdf.helpers import safe_symlink
+
+log = logging.getLogger(__name__)
 
 DEFAULT_JPEG_QUALITY = 75
 DEFAULT_PNG_QUALITY = 70
 
 
-def img_name(root, xref, ext):
-    return fspath(root / f'{xref:08d}{ext}')
+Xref = NewType('Xref', int)
 
 
-def png_name(root, xref):
+class XrefExt(NamedTuple):
+    xref: Xref
+    ext: str
+
+
+def img_name(root: Path, xref: Xref, ext: str) -> Path:
+    return root / f'{xref:08d}{ext}'
+
+
+def png_name(root: Path, xref: Xref) -> Path:
     return img_name(root, xref, '.png')
 
 
-def jpg_name(root, xref):
+def jpg_name(root: Path, xref: Xref) -> Path:
     return img_name(root, xref, '.jpg')
 
 
-def tif_name(root, xref):
+def tif_name(root: Path, xref: Xref) -> Path:
     return img_name(root, xref, '.tif')
 
 
-def extract_image_filter(pike, root, log, image, xref):
+def extract_image_filter(
+    pike: Pdf, root: Path, image: Object, xref: Xref
+) -> Optional[Tuple[PdfImage, Tuple[Name, Object]]]:
     if image.Subtype != Name.Image:
         return None
     if image.Length < 100:
         log.debug("Skipping small image, xref %s", xref)
         return None
 
-    pim = pikepdf.PdfImage(image)
+    pim = PdfImage(image)
 
     if len(pim.filter_decodeparms) > 1:
         log.debug("Skipping multiply filtered, xref %s", xref)
@@ -79,30 +107,34 @@ def extract_image_filter(pike, root, log, image, xref):
     return pim, filtdp
 
 
-def extract_image_jbig2(*, pike, root, log, image, xref, options):
-    result = extract_image_filter(pike, root, log, image, xref)
+def extract_image_jbig2(
+    *, pike: pikepdf.Pdf, root: Path, image: Object, xref: Xref, options
+) -> Optional[XrefExt]:
+    result = extract_image_filter(pike, root, image, xref)
     if result is None:
         return None
     pim, filtdp = result
 
     if (
         pim.bits_per_component == 1
-        and filtdp != Name.JBIG2Decode
+        and filtdp[0] != Name.JBIG2Decode
         and jbig2enc.available()
     ):
         try:
-            imgname = Path(root / f'{xref:08d}')
+            imgname = root / f'{xref:08d}'
             with imgname.open('wb') as f:
                 ext = pim.extract_to(stream=f)
             imgname.rename(imgname.with_suffix(ext))
         except pikepdf.UnsupportedImageTypeError:
             return None
-        return xref, ext
+        return XrefExt(xref, ext)
     return None
 
 
-def extract_image_generic(*, pike, root, log, image, xref, options):
-    result = extract_image_filter(pike, root, log, image, xref)
+def extract_image_generic(
+    *, pike: Pdf, root: Path, image: PdfImage, xref: Xref, options
+) -> Optional[XrefExt]:
+    result = extract_image_filter(pike, root, image, xref)
     if result is None:
         return None
     pim, filtdp = result
@@ -134,13 +166,13 @@ def extract_image_generic(*, pike, root, log, image, xref, options):
         #     with Image.open(stream) as im:
         #         im.save(jpg_name(root, xref), icc_profile=iccbytes)
         try:
-            imgname = Path(root / f'{xref:08d}')
+            imgname = root / f'{xref:08d}'
             with imgname.open('wb') as f:
                 ext = pim.extract_to(stream=f)
             imgname.rename(imgname.with_suffix(ext))
         except pikepdf.UnsupportedImageTypeError:
             return None
-        return xref, ext
+        return XrefExt(xref, ext)
     elif (
         pim.indexed
         and pim.colorspace in pim.SIMPLE_COLORSPACES
@@ -149,17 +181,30 @@ def extract_image_generic(*, pike, root, log, image, xref, options):
         # Try to improve on indexed images - these are far from low hanging
         # fruit in most cases
         pim.as_pil_image().save(png_name(root, xref))
-        return xref, '.png'
+        return XrefExt(xref, '.png')
     elif not pim.indexed and pim.colorspace in pim.SIMPLE_COLORSPACES:
         # An optimization opportunity here, not currently taken, is directly
         # generating a PNG from compressed data
         pim.as_pil_image().save(png_name(root, xref))
-        return xref, '.png'
+        return XrefExt(xref, '.png')
+    elif (
+        not pim.indexed
+        and pim.colorspace == Name.ICCBased
+        and pim.bits_per_component == 1
+        and not options.jbig2_lossy
+    ):
+        # We can losslessly optimize 1-bit images to CCITT or JBIG2 without
+        # paying any attention to the ICC profile, provided we're not doing
+        # lossy JBIG2
+        pim.as_pil_image().save(png_name(root, xref))
+        return XrefExt(xref, '.png')
 
     return None
 
 
-def extract_images(pike, root, log, options, extract_fn):
+def extract_images(
+    pike: Pdf, root: Path, options, extract_fn: Callable[..., Optional[XrefExt]],
+) -> Iterator[Tuple[int, XrefExt]]:
     """Extract image using extract_fn
 
     Enumerate images on each page, lookup their xref/ID number in the PDF.
@@ -175,8 +220,8 @@ def extract_images(pike, root, log, options, extract_fn):
     extension. extract_fn must also extract the file it finds interesting.
     """
 
-    include_xrefs = set()
-    exclude_xrefs = set()
+    include_xrefs: MutableSet[Xref] = set()
+    exclude_xrefs: MutableSet[Xref] = set()
     pageno_for_xref = {}
     errors = 0
     for pageno, page in enumerate(pike.pages):
@@ -187,10 +232,10 @@ def extract_images(pike, root, log, options, extract_fn):
         for _imname, image in dict(xobjs).items():
             if image.objgen[1] != 0:
                 continue  # Ignore images in an incremental PDF
-            xref = image.objgen[0]
+            xref = Xref(image.objgen[0])
             if hasattr(image, 'SMask'):
                 # Ignore soft masks
-                smask_xref = image.SMask.objgen[0]
+                smask_xref = Xref(image.SMask.objgen[0])
                 exclude_xrefs.add(smask_xref)
             include_xrefs.add(xref)
             if xref not in pageno_for_xref:
@@ -201,41 +246,41 @@ def extract_images(pike, root, log, options, extract_fn):
         image = pike.get_object((xref, 0))
         try:
             result = extract_fn(
-                pike=pike, root=root, log=log, image=image, xref=xref, options=options
+                pike=pike, root=root, image=image, xref=xref, options=options
             )
-        except Exception as e:
+        except Exception as e:  # pylint: disable=broad-except
             log.debug("Image xref %s, error %s", xref, repr(e))
             errors += 1
         else:
             if result:
                 _, ext = result
-                yield pageno_for_xref[xref], xref, ext
+                yield pageno_for_xref[xref], XrefExt(xref, ext)
 
 
-def extract_images_generic(pike, root, log, options):
+def extract_images_generic(
+    pike: Pdf, root: Path, options
+) -> Tuple[List[Xref], List[Xref]]:
     """Extract any >=2bpp image we think we can improve"""
 
     jpegs = []
     pngs = []
-    for _, xref, ext in extract_images(pike, root, log, options, extract_image_generic):
-        log.debug('xref = %s ext = %s', xref, ext)
-        if ext == '.png':
-            pngs.append(xref)
-        elif ext == '.jpg':
-            jpegs.append(xref)
+    for _, xref_ext in extract_images(pike, root, options, extract_image_generic):
+        log.debug('%s', xref_ext)
+        if xref_ext.ext == '.png':
+            pngs.append(xref_ext.xref)
+        elif xref_ext.ext == '.jpg':
+            jpegs.append(xref_ext.xref)
     log.debug("Optimizable images: JPEGs: %s PNGs: %s", len(jpegs), len(pngs))
     return jpegs, pngs
 
 
-def extract_images_jbig2(pike, root, log, options):
+def extract_images_jbig2(pike: Pdf, root: Path, options) -> Dict[int, List[XrefExt]]:
     """Extract any bitonal image that we think we can improve as JBIG2"""
 
     jbig2_groups = defaultdict(list)
-    for pageno, xref, ext in extract_images(
-        pike, root, log, options, extract_image_jbig2
-    ):
+    for pageno, xref_ext in extract_images(pike, root, options, extract_image_jbig2):
         group = pageno // options.jbig2_page_group_size
-        jbig2_groups[group].append((xref, ext))
+        jbig2_groups[group].append(xref_ext)
 
     # Elide empty groups
     jbig2_groups = {
@@ -245,55 +290,59 @@ def extract_images_jbig2(pike, root, log, options):
     return jbig2_groups
 
 
-def _produce_jbig2_images(jbig2_groups, root, log, options):
+def _produce_jbig2_images(
+    jbig2_groups: Dict[int, List[XrefExt]], root: Path, options
+) -> None:
     """Produce JBIG2 images from their groups"""
 
-    def jbig2_group_futures(executor, root, groups):
+    def jbig2_group_args(root: Path, groups: Dict[int, List[XrefExt]]):
         for group, xref_exts in groups.items():
             prefix = f'group{group:08d}'
-            future = executor.submit(
-                jbig2enc.convert_group,
+            yield dict(
                 cwd=fspath(root),
                 infiles=(img_name(root, xref, ext) for xref, ext in xref_exts),
                 out_prefix=prefix,
             )
-            yield future
 
-    def jbig2_single_futures(executor, root, groups):
+    def jbig2_single_args(root, groups: Dict[int, List[XrefExt]]):
         for group, xref_exts in groups.items():
             prefix = f'group{group:08d}'
             # Second loop is to ensure multiple images per page are unpacked
             for n, xref_ext in enumerate(xref_exts):
                 xref, ext = xref_ext
-                future = executor.submit(
-                    jbig2enc.convert_single,
+                yield dict(
                     cwd=fspath(root),
                     infile=img_name(root, xref, ext),
                     outfile=root / f'{prefix}.{n:04d}',
                 )
-                yield future
+
+    def convert_generic(fn, kwargs_dict):
+        return fn(**kwargs_dict)
 
     if options.jbig2_page_group_size > 1:
-        jbig2_futures = jbig2_group_futures
+        jbig2_args = jbig2_group_args
+        jbig2_convert = partial(convert_generic, jbig2enc.convert_group)
     else:
-        jbig2_futures = jbig2_single_futures
+        jbig2_args = jbig2_single_args
+        jbig2_convert = partial(convert_generic, jbig2enc.convert_single)
 
-    with concurrent.futures.ThreadPoolExecutor(max_workers=options.jobs) as executor:
-        futures = jbig2_futures(executor, root, jbig2_groups)
-        with tqdm(
+    exec_progress_pool(
+        use_threads=True,
+        max_workers=options.jobs,
+        tqdm_kwargs=dict(
             total=len(jbig2_groups),
             desc="JBIG2",
             unit='item',
             disable=not options.progress_bar,
-        ) as pbar:
-            for future in concurrent.futures.as_completed(futures):
-                proc = future.result()
-                if proc.stderr:
-                    log.debug(proc.stderr.decode())
-                pbar.update()
+        ),
+        task=jbig2_convert,
+        task_arguments=jbig2_args(root, jbig2_groups),
+    )
 
 
-def convert_to_jbig2(pike, jbig2_groups, root, log, options):
+def convert_to_jbig2(
+    pike: Pdf, jbig2_groups: Dict[int, List[XrefExt]], root: Path, options
+) -> None:
     """Convert images to JBIG2 and insert into PDF.
 
     When the JBIG2 page group size is > 1 we do several JBIG2 images at once
@@ -307,7 +356,7 @@ def convert_to_jbig2(pike, jbig2_groups, root, log, options):
     and needs no dictionary. Currently this must be lossless JBIG2.
     """
 
-    _produce_jbig2_images(jbig2_groups, root, log, options)
+    _produce_jbig2_images(jbig2_groups, root, options)
 
     for group, xref_exts in jbig2_groups.items():
         prefix = f'group{group:08d}'
@@ -331,19 +380,19 @@ def convert_to_jbig2(pike, jbig2_groups, root, log, options):
             )
 
 
-def transcode_jpegs(pike, jpegs, root, log, options):
+def transcode_jpegs(pike: Pdf, jpegs: Sequence[Xref], root: Path, options) -> None:
     for xref in tqdm(
         jpegs, desc="JPEGs", unit='image', disable=not options.progress_bar
     ):
-        in_jpg = Path(jpg_name(root, xref))
+        in_jpg = jpg_name(root, xref)
         opt_jpg = in_jpg.with_suffix('.opt.jpg')
 
         # This produces a debug warning from PIL
         # DEBUG:PIL.Image:Error closing: 'NoneType' object has no attribute
         # 'close'.  Seems to be mostly harmless
         # https://github.com/python-pillow/Pillow/issues/1144
-        with Image.open(fspath(in_jpg)) as im:
-            im.save(fspath(opt_jpg), optimize=True, quality=options.jpeg_quality)
+        with Image.open(in_jpg) as im:
+            im.save(opt_jpg, optimize=True, quality=options.jpeg_quality)
 
         if opt_jpg.stat().st_size > in_jpg.stat().st_size:
             log.debug("xref %s, jpeg, made larger - skip", xref)
@@ -354,37 +403,46 @@ def transcode_jpegs(pike, jpegs, root, log, options):
         im_obj.write(compdata.read(), filter=Name.DCTDecode)
 
 
-def transcode_pngs(pike, images, image_name_fn, root, log, options):
-    modified = set()
+def transcode_pngs(
+    pike: Pdf,
+    images: Sequence[Xref],
+    image_name_fn: Callable[[Path, Xref], Path],
+    root: Path,
+    options,
+) -> None:
+    modified: MutableSet[Xref] = set()
     if options.optimize >= 2:
         png_quality = (
             max(10, options.png_quality - 10),
             min(100, options.png_quality + 10),
         )
-        with concurrent.futures.ThreadPoolExecutor(
-            max_workers=options.jobs
-        ) as executor:
-            futures = []
+
+        def pngquant_args():
             for xref in images:
                 log.debug(image_name_fn(root, xref))
-                futures.append(
-                    executor.submit(
-                        pngquant.quantize,
-                        image_name_fn(root, xref),
-                        png_name(root, xref),
-                        png_quality[0],
-                        png_quality[1],
-                    )
+                yield (
+                    image_name_fn(root, xref),
+                    png_name(root, xref),
+                    png_quality[0],
+                    png_quality[1],
                 )
                 modified.add(xref)
-            with tqdm(
+
+        def pngquant_fn(args):
+            pngquant.quantize(*args)
+
+        exec_progress_pool(
+            use_threads=True,
+            max_workers=options.jobs,
+            tqdm_kwargs=dict(
                 desc="PNGs",
-                total=len(futures),
+                total=len(images),
                 unit='image',
                 disable=not options.progress_bar,
-            ) as pbar:
-                for _future in concurrent.futures.as_completed(futures):
-                    pbar.update()
+            ),
+            task=pngquant_fn,
+            task_arguments=pngquant_args(),
+        )
 
     for xref in modified:
         im_obj = pike.get_object(xref, 0)
@@ -410,12 +468,12 @@ def transcode_pngs(pike, images, image_name_fn, root, log, options):
             )
             continue
         if compdata.type == leptonica.lept.L_FLATE_ENCODE:
-            return rewrite_png(pike, im_obj, compdata, log)
+            rewrite_png(pike, im_obj, compdata)
         elif compdata.type == leptonica.lept.L_G4_ENCODE:
-            return rewrite_png_as_g4(pike, im_obj, compdata, log)
+            rewrite_png_as_g4(pike, im_obj, compdata)
 
 
-def rewrite_png_as_g4(pike, im_obj, compdata, log):
+def rewrite_png_as_g4(pike: Pdf, im_obj: Object, compdata) -> None:
     im_obj.BitsPerComponent = 1
     im_obj.Width = compdata.w
     im_obj.Height = compdata.h
@@ -435,7 +493,7 @@ def rewrite_png_as_g4(pike, im_obj, compdata, log):
     return
 
 
-def rewrite_png(pike, im_obj, compdata, log):
+def rewrite_png(pike: Pdf, im_obj: Object, compdata) -> None:
     # When a PNG is inserted into a PDF, we more or less copy the IDAT section from
     # the PDF and transfer the rest of the PNG headers to PDF image metadata.
     # One thing we have to do is tell the PDF reader whether a predictor was used
@@ -488,8 +546,7 @@ def rewrite_png(pike, im_obj, compdata, log):
     im_obj.write(compdata.read(), filter=Name.FlateDecode, decode_parms=dparms)
 
 
-def optimize(input_file, output_file, context, save_settings):
-    log = context.log
+def optimize(input_file: Path, output_file: Path, context, save_settings) -> None:
     options = context.options
     if options.optimize == 0:
         safe_symlink(input_file, output_file)
@@ -503,25 +560,25 @@ def optimize(input_file, output_file, context, save_settings):
         options.jbig2_page_group_size = 10 if options.jbig2_lossy else 1
 
     with pikepdf.Pdf.open(input_file) as pike:
-        root = Path(output_file).parent / 'images'
+        root = output_file.parent / 'images'
         root.mkdir(exist_ok=True)
 
-        jpegs, pngs = extract_images_generic(pike, root, log, options)
-        transcode_jpegs(pike, jpegs, root, log, options)
+        jpegs, pngs = extract_images_generic(pike, root, options)
+        transcode_jpegs(pike, jpegs, root, options)
         # if options.optimize >= 2:
         # Try pngifying the jpegs
-        #    transcode_pngs(pike, jpegs, jpg_name, root, log, options)
-        transcode_pngs(pike, pngs, png_name, root, log, options)
+        #    transcode_pngs(pike, jpegs, jpg_name, root, options)
+        transcode_pngs(pike, pngs, png_name, root, options)
 
-        jbig2_groups = extract_images_jbig2(pike, root, log, options)
-        convert_to_jbig2(pike, jbig2_groups, root, log, options)
+        jbig2_groups = extract_images_jbig2(pike, root, options)
+        convert_to_jbig2(pike, jbig2_groups, root, options)
 
-        target_file = Path(output_file).with_suffix('.opt.pdf')
+        target_file = output_file.with_suffix('.opt.pdf')
         pike.remove_unreferenced_resources()
         pike.save(target_file, **save_settings)
 
-    input_size = Path(input_file).stat().st_size
-    output_size = Path(target_file).stat().st_size
+    input_size = input_file.stat().st_size
+    output_size = target_file.stat().st_size
     if output_size == 0:
         raise OutputFileAccessError(
             f"Output file not created after optimizing. We probably ran "
@@ -542,8 +599,8 @@ def optimize(input_file, output_file, context, save_settings):
 
 
 def main(infile, outfile, level, jobs=1):
-    from tempfile import TemporaryDirectory
-    from shutil import copy
+    from shutil import copy  # pylint: disable=import-outside-toplevel
+    from tempfile import TemporaryDirectory  # pylint: disable=import-outside-toplevel
 
     class OptimizeOptions:
         """Emulate ocrmypdf's options"""
@@ -561,6 +618,7 @@ def main(infile, outfile, level, jobs=1):
             self.quiet = True
             self.progress_bar = False
 
+    infile = Path(infile)
     options = OptimizeOptions(
         input_file=infile,
         jobs=jobs,
@@ -571,7 +629,7 @@ def main(infile, outfile, level, jobs=1):
     )
 
     with TemporaryDirectory() as td:
-        context = PDFContext(options, td, infile, None)
+        context = PdfContext(options, td, infile, None, None)
         tmpout = Path(td) / 'out.pdf'
         optimize(
             infile,
