@@ -16,4 +16,4 @@
 # You should have received a copy of the GNU General Public License
 # along with OCRmyPDF.  If not, see <http://www.gnu.org/licenses/>.
 
-from .info import Colorspace, Encoding, PdfInfo
+from ocrmypdf.pdfinfo.info import Colorspace, Encoding, PdfInfo
