@@ -18,29 +18,30 @@
 
 import logging
 import re
-from collections import namedtuple
+from collections import defaultdict, namedtuple
 from decimal import Decimal
 from enum import Enum
+from functools import partial
 from math import hypot, isclose
-from os import fspath
+from os import PathLike
 from pathlib import Path
+from typing import Any, Dict, List, Optional, Union
 from warnings import warn
 
 import pikepdf
 from pikepdf import PdfMatrix
-from tqdm import tqdm
 
-from ocrmypdf.exceptions import EncryptedPdfError, MissingDependencyError
-
-from . import ghosttext
-from .layout import get_page_analysis, get_text_boxes
+from ocrmypdf._concurrent import exec_progress_pool
+from ocrmypdf.exceptions import EncryptedPdfError
+from ocrmypdf.helpers import Resolution, available_cpu_count, pikepdf_enable_mmap
+from ocrmypdf.pdfinfo.layout import get_page_analysis, get_text_boxes
 
 logger = logging.getLogger()
 
 Colorspace = Enum('Colorspace', 'gray rgb cmyk lab icc index sep devn pattern jpeg2000')
 
 Encoding = Enum(
-    'Encoding', 'ccitt jpeg jpeg2000 jbig2 asciihex ascii85 lzw flate ' + 'runlength'
+    'Encoding', 'ccitt jpeg jpeg2000 jbig2 asciihex ascii85 lzw flate runlength'
 )
 
 FRIENDLY_COLORSPACE = {
@@ -90,7 +91,7 @@ UNIT_SQUARE = (1.0, 0.0, 0.0, 1.0, 0.0, 0.0)
 def _is_unit_square(shorthand):
     values = map(float, shorthand)
     pairwise = zip(values, UNIT_SQUARE)
-    return all([isclose(a, b, rel_tol=1e-3) for a, b in pairwise])
+    return all(isclose(a, b, rel_tol=1e-3) for a, b in pairwise)
 
 
 XobjectSettings = namedtuple('XobjectSettings', ['name', 'shorthand', 'stack_depth'])
@@ -98,15 +99,19 @@ XobjectSettings = namedtuple('XobjectSettings', ['name', 'shorthand', 'stack_dep
 InlineSettings = namedtuple('InlineSettings', ['iimage', 'shorthand', 'stack_depth'])
 
 ContentsInfo = namedtuple(
-    'ContentsInfo', ['xobject_settings', 'inline_images', 'found_vector']
+    'ContentsInfo',
+    ['xobject_settings', 'inline_images', 'found_vector', 'found_text', 'name_index'],
 )
 
 TextboxInfo = namedtuple('TextboxInfo', ['bbox', 'is_visible', 'is_corrupt'])
 
 
-class VectorInfo:
-    def __init__(self):
-        pass
+class VectorMarker:
+    pass
+
+
+class TextMarker:
+    pass
 
 
 def _normalize_stack(graphobjs):
@@ -151,10 +156,13 @@ def _interpret_contents(contentstream, initial_shorthand=UNIT_SQUARE):
     ctm = PdfMatrix(initial_shorthand)
     xobject_settings = []
     inline_images = []
+    name_index = defaultdict(lambda: [])
     found_vector = False
+    found_text = False
     vector_ops = set('S s f F f* B B* b b*'.split())
+    text_showing_ops = set("""TJ Tj " '""".split())
     image_ops = set('BI ID EI q Q Do cm'.split())
-    operator_whitelist = ' '.join(vector_ops | image_ops)
+    operator_whitelist = ' '.join(vector_ops | text_showing_ops | image_ops)
 
     for n, graphobj in enumerate(
         _normalize_stack(
@@ -185,6 +193,7 @@ def _interpret_contents(contentstream, initial_shorthand=UNIT_SQUARE):
                 name=image_name, shorthand=ctm.shorthand, stack_depth=len(stack)
             )
             xobject_settings.append(settings)
+            name_index[image_name].append(settings)
         elif operator == 'INLINE IMAGE':  # BI/ID/EI are grouped into this
             iimage = operands[0]
             inline = InlineSettings(
@@ -193,11 +202,15 @@ def _interpret_contents(contentstream, initial_shorthand=UNIT_SQUARE):
             inline_images.append(inline)
         elif operator in vector_ops:
             found_vector = True
+        elif operator in text_showing_ops:
+            found_text = True
 
     return ContentsInfo(
         xobject_settings=xobject_settings,
         inline_images=inline_images,
         found_vector=found_vector,
+        found_text=found_text,
+        name_index=name_index,
     )
 
 
@@ -224,7 +237,7 @@ def _get_dpi(ctm_shorthand, image_size):
     width-axis vector v0 (1, 0), height-axis vector vh (0, 1) with the matrix,
     which gives the dimensions of the image in PDF units. From there we can
     compare to actual image dimensions. PDF uses
-    row vector * matrix_tranposed unlike the traditional
+    row vector * matrix_transposed unlike the traditional
     matrix * column vector.
 
     The offset, width and height vectors can be combined in a matrix and
@@ -262,7 +275,7 @@ def _get_dpi(ctm_shorthand, image_size):
     dpi_w = scale_w * 72.0
     dpi_h = scale_h * 72.0
 
-    return dpi_w, dpi_h
+    return Resolution(dpi_w, dpi_h)
 
 
 class ImageInfo:
@@ -303,14 +316,22 @@ class ImageInfo:
         if self._enc == Encoding.jpeg2000:
             self._color = Colorspace.jpeg2000
 
-        self._comp = FRIENDLY_COMP.get(self._color, '?')
+        if self._color == Colorspace.icc:
+            # Check the ICC profile to determine actual colorspace
+            pim_icc = pim.icc
+            if pim_icc.profile.xcolor_space == 'GRAY':
+                self._comp = 1
+            elif pim_icc.profile.xcolor_space == 'CMYK':
+                self._comp = 4
+            else:
+                self._comp = 3
+        else:
+            self._comp = FRIENDLY_COMP.get(self._color, '?')
 
-        # Bit of a hack... infer grayscale if component count is uncertain
-        # but encoding must be monochrome. This happens if a monochrome image
-        # has an ICC profile attached. Better solution would be to examine
-        # the ICC profile.
-        if self._comp == '?' and self._enc in (Encoding.ccitt, Encoding.jbig2):
-            self._comp = FRIENDLY_COMP[Colorspace.gray]
+            # Bit of a hack... infer grayscale if component count is uncertain
+            # but encoding only supports monochrome.
+            if self._comp == '?' and self._enc in (Encoding.ccitt, Encoding.jbig2):
+                self._comp = FRIENDLY_COMP[Colorspace.gray]
 
     @property
     def name(self):
@@ -345,12 +366,8 @@ class ImageInfo:
         return self._enc
 
     @property
-    def xres(self):
-        return _get_dpi(self._shorthand, (self._width, self._height))[0]
-
-    @property
-    def yres(self):
-        return _get_dpi(self._shorthand, (self._width, self._height))[1]
+    def dpi(self):
+        return _get_dpi(self._shorthand, (self._width, self._height))
 
     def __repr__(self):
         class_locals = {
@@ -360,7 +377,7 @@ class ImageInfo:
         }
         return (
             "<ImageInfo '{name}' {type_} {width}x{height} {color} "
-            "{comp} {bpc} {enc} {xres}x{yres}>"
+            "{comp} {bpc} {enc} {dpi}>"
         ).format(**class_locals)
 
 
@@ -411,13 +428,9 @@ def _find_regular_images(container, contentsinfo):
     """
 
     for pdfimage, xobj in _image_xobjects(container):
-
-        # For each image that is drawn on this, check if we drawing the
-        # current image - yes this is O(n^2), but n == 1 almost always
-        for draw in contentsinfo.xobject_settings:
-            if draw.name != xobj:
-                continue
-
+        if xobj not in contentsinfo.name_index:
+            continue
+        for draw in contentsinfo.name_index[xobj]:
             if draw.stack_depth == 0 and _is_unit_square(draw.shorthand):
                 # At least one PDF in the wild (and test suite) draws an image
                 # when the graphics stack depth is 0, meaning that the image
@@ -502,13 +515,15 @@ def _process_content_streams(*, pdf, container, shorthand=None):
     contentsinfo = _interpret_contents(container, initial_shorthand)
 
     if contentsinfo.found_vector:
-        yield VectorInfo()
+        yield VectorMarker()
+    if contentsinfo.found_text:
+        yield TextMarker()
     yield from _find_inline_images(contentsinfo)
     yield from _find_regular_images(container, contentsinfo)
     yield from _find_form_xobject_images(pdf, container, contentsinfo)
 
 
-def _page_has_text(text_blocks, page_width, page_height):
+def _page_has_text(text_blocks, page_width, page_height) -> bool:
     """Smarter text detection that ignores text in margins"""
 
     pw, ph = float(page_width), float(page_height)
@@ -551,8 +566,10 @@ def simplify_textboxes(miner, textbox_getter):
         yield TextboxInfo(box.bbox, visible, corrupt)
 
 
-def _pdf_get_pageinfo(pdf, pageno: int, infile, xmltext):
-    pageinfo = {}
+def _pdf_get_pageinfo(
+    pdf, pageno: int, infile: PathLike, check_pages, detailed_analysis: bool
+):
+    pageinfo: Dict[str, Any] = {}
     pageinfo['pageno'] = pageno
     pageinfo['images'] = []
 
@@ -561,18 +578,18 @@ def _pdf_get_pageinfo(pdf, pageno: int, infile, xmltext):
     width_pt = mediabox[2] - mediabox[0]
     height_pt = mediabox[3] - mediabox[1]
 
-    if xmltext is not None:
-        bboxes = ghosttext.page_get_textblocks(
-            fspath(infile), pageno, xmltext=xmltext, height=height_pt
-        )
-        pageinfo['bboxes'] = bboxes
-    else:
+    check_this_page = pageno in check_pages
+
+    if check_this_page and detailed_analysis:
         pscript5_mode = str(pdf.docinfo.get('/Creator')).startswith('PScript5')
         miner = get_page_analysis(infile, pageno, pscript5_mode)
         pageinfo['textboxes'] = list(simplify_textboxes(miner, get_text_boxes))
         bboxes = (box.bbox for box in pageinfo['textboxes'])
 
-    pageinfo['has_text'] = _page_has_text(bboxes, width_pt, height_pt)
+        pageinfo['has_text'] = _page_has_text(bboxes, width_pt, height_pt)
+    else:
+        pageinfo['textboxes'] = []
+        pageinfo['has_text'] = None  # i.e. "no information"
 
     userunit = page.get('/UserUnit', Decimal(1.0))
     if not isinstance(userunit, Decimal):
@@ -587,96 +604,134 @@ def _pdf_get_pageinfo(pdf, pageno: int, infile, xmltext):
         pageinfo['rotate'] = 0
 
     userunit_shorthand = (userunit, 0, 0, userunit, 0, 0)
-    contentsinfo = [
-        ci
+
+    if check_this_page:
+        pageinfo['has_vector'] = False
+        pageinfo['has_text'] = False
+        pageinfo['images'] = []
         for ci in _process_content_streams(
             pdf=pdf, container=page, shorthand=userunit_shorthand
-        )
-    ]
+        ):
+            if isinstance(ci, VectorMarker):
+                pageinfo['has_vector'] = True
+            elif isinstance(ci, TextMarker):
+                pageinfo['has_text'] = True
+            elif isinstance(ci, ImageInfo):
+                pageinfo['images'].append(ci)
+            else:
+                raise NotImplementedError()
+    else:
+        pageinfo['has_vector'] = None  # i.e. "no information"
+        pageinfo['has_text'] = None
+        pageinfo['images'] = None
 
-    pageinfo['has_vector'] = False
-    if any(isinstance(ci, VectorInfo) for ci in contentsinfo):
-        pageinfo['has_vector'] = True
-
-    pageinfo['images'] = [im for im in contentsinfo if isinstance(im, ImageInfo)]
     if pageinfo['images']:
-        xres = Decimal(max(image.xres for image in pageinfo['images']))
-        yres = Decimal(max(image.yres for image in pageinfo['images']))
-        pageinfo['xres'], pageinfo['yres'] = xres, yres
-        pageinfo['width_pixels'] = int(round(xres * pageinfo['width_inches']))
-        pageinfo['height_pixels'] = int(round(yres * pageinfo['height_inches']))
+        dpi = Resolution(0.0, 0.0).take_max(image.dpi for image in pageinfo['images'])
+        pageinfo['dpi'] = dpi
+        pageinfo['width_pixels'] = int(round(dpi.x * float(pageinfo['width_inches'])))
+        pageinfo['height_pixels'] = int(round(dpi.y * float(pageinfo['height_inches'])))
 
     return pageinfo
 
 
-def _pdf_get_all_pageinfo(infile, detailed_analysis=False, log=None, progbar=False):
-    pdf = pikepdf.open(infile)  # Do not close in this function
-    if pdf.is_encrypted:
-        pdf.close()
-        raise EncryptedPdfError()  # Triggered by encryption with empty passwd
-    if detailed_analysis:
-        pages_xml = None
-    else:
-        pages_xml = ghosttext.extract_text_xml(infile, pdf, pageno=None, log=log)
+worker_pdf = None
 
-    pages = []
-    for n, _ in tqdm(
-        enumerate(pdf.pages),
-        total=len(pdf.pages),
-        desc="Scan",
-        unit='page',
-        disable=not progbar,
-    ):
-        page_xml = pages_xml[n] if pages_xml else None
-        page = PageInfo(pdf, n, infile, page_xml, detailed_analysis)
-        pages.append(page)
 
-    return pages, pdf
+def _pdf_pageinfo_sync_init(infile):
+    global worker_pdf  # pylint: disable=global-statement
+    pikepdf_enable_mmap()
+    worker_pdf = pikepdf.open(infile)
+
+
+def _pdf_pageinfo_sync(args):
+    global worker_pdf  # pylint: disable=global-statement
+    pageno, infile, check_pages, detailed_analysis = args
+    page = PageInfo(worker_pdf, pageno, infile, check_pages, detailed_analysis)
+    return page
+
+
+def _pdf_pageinfo_concurrent(
+    pdf, infile, progbar, max_workers, check_pages, detailed_analysis=False
+):
+    pages = [None] * len(pdf.pages)
+
+    def update_pageinfo(result, pbar):
+        page = result
+        pages[page.pageno] = page
+        pbar.update()
+
+    if max_workers is None:
+        max_workers = available_cpu_count()
+
+    total = len(pdf.pages)
+    contexts = ((n, infile, check_pages, detailed_analysis) for n in range(total))
+
+    use_threads = False  # No performance gain if threaded due to GIL
+    n_workers = min(1 + len(pages) // 4, max_workers)
+    if n_workers == 1:
+        # But if we decided on only one worker, there is no point in using
+        # a separate process.
+        use_threads = True
+
+    exec_progress_pool(
+        use_threads=use_threads,
+        max_workers=n_workers,
+        tqdm_kwargs=dict(
+            total=total, desc="Scanning contents", unit='page', disable=not progbar
+        ),
+        task_initializer=partial(_pdf_pageinfo_sync_init, infile),
+        task=_pdf_pageinfo_sync,
+        task_arguments=contexts,
+        task_finished=update_pageinfo,
+    )
+    return pages
 
 
 class PageInfo:
-    def __init__(self, pdf, pageno, infile, xmltext, detailed_analysis=False):
+    def __init__(self, pdf, pageno, infile, check_pages, detailed_analysis=False):
         self._pageno = pageno
         self._infile = infile
-        self._pageinfo = _pdf_get_pageinfo(pdf, pageno, infile, xmltext)
         self._detailed_analysis = detailed_analysis
+        self._pageinfo = _pdf_get_pageinfo(
+            pdf, pageno, infile, check_pages, detailed_analysis
+        )
 
     @property
-    def pageno(self):
+    def pageno(self) -> int:
         return self._pageno
 
     @property
-    def has_text(self):
+    def has_text(self) -> bool:
         return self._pageinfo['has_text']
 
     @property
-    def has_corrupt_text(self):
+    def has_corrupt_text(self) -> bool:
         if not self._detailed_analysis:
             raise NotImplementedError('Did not do detailed analysis')
         return any(tbox.is_corrupt for tbox in self._pageinfo['textboxes'])
 
     @property
-    def has_vector(self):
+    def has_vector(self) -> bool:
         return self._pageinfo['has_vector']
 
     @property
-    def width_inches(self):
+    def width_inches(self) -> Decimal:
         return self._pageinfo['width_inches']
 
     @property
-    def height_inches(self):
+    def height_inches(self) -> Decimal:
         return self._pageinfo['height_inches']
 
     @property
-    def width_pixels(self):
-        return int(round(self.width_inches * self.xres))
+    def width_pixels(self) -> int:
+        return int(round(float(self.width_inches) * self.dpi.x))
 
     @property
-    def height_pixels(self):
-        return int(round(self.height_inches * self.yres))
+    def height_pixels(self) -> int:
+        return int(round(float(self.height_inches) * self.dpi.y))
 
     @property
-    def rotation(self):
+    def rotation(self) -> int:
         return self._pageinfo.get('rotate', None)
 
     @rotation.setter
@@ -690,7 +745,9 @@ class PageInfo:
     def images(self):
         return self._pageinfo['images']
 
-    def get_textareas(self, visible=None, corrupt=None):
+    def get_textareas(
+        self, visible: Optional[bool] = None, corrupt: Optional[bool] = None
+    ):
         def predicate(obj, want_visible, want_corrupt):
             result = True
             if want_visible is not None:
@@ -703,7 +760,7 @@ class PageInfo:
 
         if 'textboxes' not in self._pageinfo:
             if visible is not None and corrupt is not None:
-                raise NotImplementedError('Ghostscript textboxes cannot be classified')
+                raise NotImplementedError('Incomplete information on textboxes')
             return self._pageinfo['bboxes']
 
         return (
@@ -713,19 +770,15 @@ class PageInfo:
         )
 
     @property
-    def xres(self):
-        return self._pageinfo.get('xres', None)
+    def dpi(self) -> Resolution:
+        return self._pageinfo.get('dpi', Resolution(0.0, 0.0))
 
     @property
-    def yres(self):
-        return self._pageinfo.get('yres', None)
-
-    @property
-    def userunit(self):
+    def userunit(self) -> Decimal:
         return self._pageinfo.get('userunit', None)
 
     @property
-    def min_version(self):
+    def min_version(self) -> str:
         if self.userunit is not None:
             return '1.6'
         else:
@@ -733,63 +786,74 @@ class PageInfo:
 
     def __repr__(self):
         return (
-            '<PageInfo ' 'pageno={} {}"x{}" rotation={} res={}x{} has_text={}>'
-        ).format(
-            self.pageno,
-            self.width_inches,
-            self.height_inches,
-            self.rotation,
-            self.xres,
-            self.yres,
-            self.has_text,
+            f'<PageInfo '
+            f'pageno={self.pageno} {self.width_inches}"x{self.height_inches}" '
+            f'rotation={self.rotation} dpi={self.dpi} has_text={self.has_text}>'
         )
 
 
 class PdfInfo:
     """Get summary information about a PDF"""
 
-    def __init__(self, infile, detailed_page_analysis=False, log=logger, progbar=False):
+    def __init__(
+        self,
+        infile,
+        detailed_analysis: bool = False,
+        progbar: bool = False,
+        max_workers: int = None,
+        check_pages=None,
+    ):
         self._infile = infile
-        self._pages, pdf = _pdf_get_all_pageinfo(
-            infile, detailed_page_analysis, log=log, progbar=progbar
-        )
-        self._needs_rendering = pdf.root.get('/NeedsRendering', False)
-        self._has_acroform = False
-        if '/AcroForm' in pdf.root:
-            if len(pdf.root.AcroForm.get('/Fields', [])) > 0:
-                self._has_acroform = True
-            elif '/XFA' in pdf.root.AcroForm:
-                self._has_acroform = True
-        pdf.close()
+        if check_pages is None:
+            check_pages = range(0, 1_000_000_000)
+
+        with pikepdf.open(infile) as pdf:
+            if pdf.is_encrypted:
+                raise EncryptedPdfError()  # Triggered by encryption with empty passwd
+            self._pages = _pdf_pageinfo_concurrent(
+                pdf,
+                infile,
+                progbar,
+                max_workers,
+                check_pages=check_pages,
+                detailed_analysis=detailed_analysis,
+            )
+            self._needs_rendering = pdf.root.get('/NeedsRendering', False)
+            self._has_acroform = False
+            if '/AcroForm' in pdf.root:
+                if len(pdf.root.AcroForm.get('/Fields', [])) > 0:
+                    self._has_acroform = True
+                elif '/XFA' in pdf.root.AcroForm:
+                    self._has_acroform = True
 
     @property
     def pages(self):
         return self._pages
 
     @property
-    def min_version(self):
+    def min_version(self) -> str:
         # The minimum PDF is the maximum version that any particular page needs
         return max(page.min_version for page in self.pages)
 
     @property
-    def has_userunit(self):
+    def has_userunit(self) -> bool:
         return any(page.userunit != 1.0 for page in self.pages)
 
     @property
-    def has_acroform(self):
+    def has_acroform(self) -> bool:
         return self._has_acroform
 
     @property
-    def filename(self):
+    def filename(self) -> Union[str, Path]:
         if not isinstance(self._infile, (str, Path)):
             raise NotImplementedError("can't get filename from stream")
         return self._infile
 
     @property
-    def needs_rendering(self):
+    def needs_rendering(self) -> bool:
         return self._needs_rendering
 
-    def __getitem__(self, item):
+    def __getitem__(self, item) -> PageInfo:
         return self._pages[item]
 
     def __len__(self):
@@ -800,15 +864,19 @@ class PdfInfo:
 
 
 def main():
-    import argparse
+    import argparse  # pylint: disable=import-outside-toplevel
+    from pprint import pprint  # pylint: disable=import-outside-toplevel
 
     parser = argparse.ArgumentParser()
     parser.add_argument('infile')
     args = parser.parse_args()
-    info = _pdf_get_all_pageinfo(args.infile)
-    from pprint import pprint
+    pdfinfo = PdfInfo(args.infile)
 
-    pprint(info)
+    pprint(pdfinfo)
+    for page in pdfinfo.pages:
+        pprint(page)
+        for im in page.images:
+            pprint(im)
 
 
 if __name__ == '__main__':
