@@ -15,7 +15,6 @@
 # You should have received a copy of the GNU General Public License
 # along with OCRmyPDF.  If not, see <http://www.gnu.org/licenses/>.
 
-import ast
 import os
 import platform
 import sys
@@ -24,7 +23,9 @@ from subprocess import PIPE, run
 
 import pytest
 
-from ocrmypdf import api, cli
+from ocrmypdf import api, pdfinfo
+from ocrmypdf._exec import unpaper
+from ocrmypdf._plugin_manager import get_parser_options_plugins
 
 pytest_plugins = ['helpers_namespace']
 
@@ -51,7 +52,7 @@ def is_macos():
 def running_in_docker():
     # Docker creates a file named /.dockerenv (newer versions) or
     # /.dockerinit (older) -- this is undocumented, not an offical test
-    return os.path.exists('/.dockerenv') or os.path.exists('/.dockerinit')
+    return Path('/.dockerenv').exists() or Path('/.dockerinit').exists()
 
 
 @pytest.helpers.register
@@ -62,87 +63,17 @@ def running_in_travis():
 @pytest.helpers.register
 def have_unpaper():
     try:
-        from ocrmypdf.exec import unpaper
-
         unpaper.version()
-    except Exception:
+    except Exception:  # pylint: disable=broad-except
         return False
     return True
 
 
-TESTS_ROOT = os.path.abspath(os.path.dirname(__file__))
-SPOOF_PATH = os.path.join(TESTS_ROOT, 'spoof')
-PROJECT_ROOT = os.path.dirname(TESTS_ROOT)
+TESTS_ROOT = Path(__file__).parent.resolve()
+PROJECT_ROOT = TESTS_ROOT
 OCRMYPDF = [sys.executable, '-m', 'ocrmypdf']
 
 
-WINDOWS_SHIM_TEMPLATE = """
-# This is a shim for Windows that has the same effect as a symlink to the target .py
-# file
-import os
-import subprocess
-import sys
-
-args = [sys.executable, {spoofer}, *sys.argv[1:]]
-p = subprocess.run(args, check=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-sys.stdout.buffer.write(p.stdout)
-sys.stderr.buffer.write(p.stderr)
-sys.exit(p.returncode)
-"""
-
-assert ast.parse(WINDOWS_SHIM_TEMPLATE.format(spoofer=repr(r"C:\\Temp\\file.py")))
-
-
-@pytest.helpers.register
-def spoof(tmp_path_factory, **kwargs):
-    """Modify PATH to override subprocess executables
-
-    spoof(program1='replacement', ...)
-
-    Creates temporary directory with symlinks to targets.
-
-    """
-    env = os.environ.copy()
-    slug = '-'.join(v.replace('.py', '') for v in sorted(kwargs.values()))
-    spoofer_base = tmp_path_factory.mktemp('spoofers')
-    tmpdir = Path(spoofer_base / slug)
-    tmpdir.mkdir(parents=True)
-
-    for replace_program, with_spoof in kwargs.items():
-        spoofer = Path(SPOOF_PATH) / with_spoof
-        if os.name != 'nt':
-            spoofer.chmod(0o755)
-            (tmpdir / replace_program).symlink_to(spoofer)
-        else:
-            py_file = WINDOWS_SHIM_TEMPLATE.format(
-                spoofer=repr(os.fspath(spoofer.absolute()))
-            )
-            if replace_program == 'gs':
-                programs = ['gswin64c', 'gswin32c']
-            else:
-                programs = [replace_program]
-            for prog in programs:
-                (tmpdir / f'{prog}.py').write_text(py_file, encoding='utf-8')
-
-    env['_OCRMYPDF_TEST_PATH'] = str(tmpdir) + os.pathsep + env['PATH']
-    if os.name == 'nt':
-        if '.py' not in env['PATHEXT'].lower():
-            raise EnvironmentError("PATHEXT is not configured to support .py")
-    return env
-
-
-@pytest.fixture
-def spoof_tesseract_noop(tmp_path_factory):
-    return spoof(tmp_path_factory, tesseract='tesseract_noop.py')
-
-
-@pytest.fixture
-def spoof_tesseract_cache(tmp_path_factory):
-    if running_in_docker():
-        return os.environ.copy()
-    return spoof(tmp_path_factory, tesseract="tesseract_cache.py")
-
-
 @pytest.fixture
 def resources():
     return Path(TESTS_ROOT) / 'resources'
@@ -173,58 +104,43 @@ def no_outpdf(tmp_path):
 
 
 @pytest.helpers.register
-def check_ocrmypdf(input_file, output_file, *args, env=None):
+def check_ocrmypdf(input_file, output_file, *args):
     """Run ocrmypdf and confirmed that a valid file was created"""
+    args = [str(input_file), str(output_file)] + [
+        str(arg) for arg in args if arg is not None
+    ]
 
-    options = cli.parser.parse_args(
-        [str(input_file), str(output_file)]
-        + [str(arg) for arg in args if arg is not None]
-    )
-    api.check_options(options)
-    if env:
-        options.tesseract_env = env
-        options.tesseract_env['_OCRMYPDF_TEST_INFILE'] = os.fspath(input_file)
-    result = api.run_pipeline(options, api=True)
+    _parser, options, plugin_manager = get_parser_options_plugins(args=args)
+    api.check_options(options, plugin_manager)
+    result = api.run_pipeline(options, plugin_manager=plugin_manager, api=True)
 
     assert result == 0
-    assert os.path.exists(str(output_file)), "Output file not created"
-    assert os.stat(str(output_file)).st_size > 100, "PDF too small or empty"
+    assert output_file.exists(), "Output file not created"
+    assert output_file.stat().st_size > 100, "PDF too small or empty"
 
     return output_file
 
 
 @pytest.helpers.register
-def run_ocrmypdf_api(input_file, output_file, *args, env=None):
+def run_ocrmypdf_api(input_file, output_file, *args):
     """Run ocrmypdf via API and let caller deal with results
 
     Does not currently have a way to manipulate the PATH except for Tesseract.
     """
 
-    options = cli.parser.parse_args(
-        [str(input_file), str(output_file)]
-        + [str(arg) for arg in args if arg is not None]
-    )
-    api.check_options(options)
-    if env:
-        options.tesseract_env = env.copy()
-        options.tesseract_env['_OCRMYPDF_TEST_INFILE'] = os.fspath(input_file)
-        first_path = env.get('_OCRMYPDF_TEST_PATH', '').split(os.pathsep)[0]
-        if 'spoof' in first_path:
-            assert 'gs' not in first_path, "use run_ocrmypdf() for gs"
-            assert 'tesseract' in first_path
-    if options.tesseract_env:
-        assert all(isinstance(v, (str, bytes)) for v in options.tesseract_env.values())
+    args = [str(input_file), str(output_file)] + [
+        str(arg) for arg in args if arg is not None
+    ]
+    _parser, options, plugin_manager = get_parser_options_plugins(args=args)
 
-    return api.run_pipeline(options, api=False)
+    api.check_options(options, plugin_manager)
+    return api.run_pipeline(options, plugin_manager=None, api=False)
 
 
 @pytest.helpers.register
-def run_ocrmypdf(input_file, output_file, *args, env=None, universal_newlines=True):
+def run_ocrmypdf(input_file, output_file, *args, universal_newlines=True):
     "Run ocrmypdf and let caller deal with results"
 
-    if env is None:
-        env = os.environ.copy()
-
     p_args = (
         OCRMYPDF
         + [str(arg) for arg in args if arg is not None]
@@ -232,14 +148,22 @@ def run_ocrmypdf(input_file, output_file, *args, env=None, universal_newlines=Tr
     )
 
     # Tell subprocess where to find coverage.py configuration
-    # This has no effect except when coverage is running
+    # This has no unless except when coverage is running
     # Details: https://coverage.readthedocs.io/en/coverage-5.0/subprocess.html
     coverage_rc = Path(__file__).parent.parent / '.coveragerc'
-    assert coverage_rc.exists()
-    env['COVERAGE_PROCESS_START'] = os.fspath(coverage_rc)
+    env = os.environ.copy()
+    if coverage_rc.exists():
+        env['COVERAGE_PROCESS_START'] = os.fspath(coverage_rc)
+    elif not running_in_docker():
+        assert False, "could not find .coveragerc"
 
     p = run(
-        p_args, stdout=PIPE, stderr=PIPE, universal_newlines=universal_newlines, env=env
+        p_args,
+        stdout=PIPE,
+        stderr=PIPE,
+        universal_newlines=universal_newlines,
+        env=env,
+        check=False,
     )
     # print(p.stderr)
     return p, p.stdout, p.stderr
@@ -247,8 +171,6 @@ def run_ocrmypdf(input_file, output_file, *args, env=None, universal_newlines=Tr
 
 @pytest.helpers.register
 def first_page_dimensions(pdf):
-    from ocrmypdf import pdfinfo
-
     info = pdfinfo.PdfInfo(pdf)
     page0 = info[0]
     return (page0.width_inches, page0.height_inches)
