@@ -22,41 +22,15 @@ import pikepdf
 import pytest
 from PIL import Image
 
+from ocrmypdf._exec.ghostscript import rasterize_pdf
 from ocrmypdf.exceptions import ExitCode
-from ocrmypdf.exec.ghostscript import rasterize_pdf
+from ocrmypdf.helpers import Resolution
 
-check_ocrmypdf = pytest.helpers.check_ocrmypdf
-run_ocrmypdf = pytest.helpers.run_ocrmypdf
-run_ocrmypdf_api = pytest.helpers.run_ocrmypdf_api
-spoof = pytest.helpers.spoof
+check_ocrmypdf = pytest.helpers.check_ocrmypdf  # pylint: disable=no-member
+run_ocrmypdf = pytest.helpers.run_ocrmypdf  # pylint: disable=no-member
+run_ocrmypdf_api = pytest.helpers.run_ocrmypdf_api  # pylint: disable=no-member
 
-
-@pytest.fixture
-def spoof_no_tess_gs_render_fail(tmp_path_factory):
-    return spoof(
-        tmp_path_factory, tesseract='tesseract_noop.py', gs='gs_render_failure.py'
-    )
-
-
-@pytest.fixture
-def spoof_no_tess_gs_raster_fail(tmp_path_factory):
-    return spoof(
-        tmp_path_factory, tesseract='tesseract_noop.py', gs='gs_raster_failure.py'
-    )
-
-
-@pytest.fixture
-def spoof_no_tess_no_pdfa(tmp_path_factory):
-    return spoof(
-        tmp_path_factory, tesseract='tesseract_noop.py', gs='gs_pdfa_failure.py'
-    )
-
-
-@pytest.fixture
-def spoof_no_tess_pdfa_warning(tmp_path_factory):
-    return spoof(
-        tmp_path_factory, tesseract='tesseract_noop.py', gs='gs_feature_elision.py'
-    )
+# pylint: disable=redefined-outer-name
 
 
 @pytest.fixture
@@ -65,22 +39,21 @@ def francais(resources):
     return path, pikepdf.open(path)
 
 
-def test_rasterize_size(francais, outdir, caplog):
+def test_rasterize_size(francais, outdir):
     path, pdf = francais
     page_size_pts = (pdf.pages[0].MediaBox[2], pdf.pages[0].MediaBox[3])
     assert pdf.pages[0].MediaBox[0] == pdf.pages[0].MediaBox[1] == 0
     page_size = (page_size_pts[0] / Decimal(72), page_size_pts[1] / Decimal(72))
     target_size = Decimal('50.0'), Decimal('30.0')
-    forced_dpi = 42.0, 4242.0
+    forced_dpi = Resolution(42.0, 4242.0)
 
-    log = logging.getLogger()
     rasterize_pdf(
         path,
         outdir / 'out.png',
-        target_size[0] / page_size[0],
-        target_size[1] / page_size[1],
         raster_device='pngmono',
-        log=log,
+        raster_dpi=Resolution(
+            target_size[0] / page_size[0], target_size[1] / page_size[1]
+        ),
         page_dpi=forced_dpi,
     )
 
@@ -95,17 +68,16 @@ def test_rasterize_rotated(francais, outdir, caplog):
     assert pdf.pages[0].MediaBox[0] == pdf.pages[0].MediaBox[1] == 0
     page_size = (page_size_pts[0] / Decimal(72), page_size_pts[1] / Decimal(72))
     target_size = Decimal('50.0'), Decimal('30.0')
-    forced_dpi = 42.0, 4242.0
+    forced_dpi = Resolution(42.0, 4242.0)
 
-    log = logging.getLogger()
     caplog.set_level(logging.DEBUG)
     rasterize_pdf(
         path,
         outdir / 'out.png',
-        target_size[0] / page_size[0],
-        target_size[1] / page_size[1],
         raster_device='pngmono',
-        log=log,
+        raster_dpi=Resolution(
+            target_size[0] / page_size[0], target_size[1] / page_size[1]
+        ),
         page_dpi=forced_dpi,
         rotation=90,
     )
@@ -115,30 +87,52 @@ def test_rasterize_rotated(francais, outdir, caplog):
         assert im.info['dpi'] == (forced_dpi[1], forced_dpi[0])
 
 
-def test_gs_render_failure(spoof_no_tess_gs_render_fail, resources, outpdf):
-    p, out, err = run_ocrmypdf(
-        resources / 'blank.pdf', outpdf, env=spoof_no_tess_gs_render_fail
+def test_gs_render_failure(resources, outpdf):
+    p, _out, err = run_ocrmypdf(
+        resources / 'blank.pdf',
+        outpdf,
+        '--plugin',
+        'tests/plugins/tesseract_noop.py',
+        '--plugin',
+        'tests/plugins/gs_render_failure.py',
     )
     assert 'Casper is not a friendly ghost' in err
     assert p.returncode == ExitCode.child_process_error
 
 
-def test_gs_raster_failure(spoof_no_tess_gs_raster_fail, resources, outpdf):
-    p, out, err = run_ocrmypdf(
-        resources / 'francais.pdf', outpdf, env=spoof_no_tess_gs_raster_fail
+def test_gs_raster_failure(resources, outpdf):
+    p, _out, err = run_ocrmypdf(
+        resources / 'francais.pdf',
+        outpdf,
+        '--plugin',
+        'tests/plugins/tesseract_noop.py',
+        '--plugin',
+        'tests/plugins/gs_raster_failure.py',
     )
     assert 'Ghost story archive not found' in err
     assert p.returncode == ExitCode.child_process_error
 
 
-def test_ghostscript_pdfa_failure(spoof_no_tess_no_pdfa, resources, outpdf):
-    p, out, err = run_ocrmypdf(
-        resources / 'francais.pdf', outpdf, env=spoof_no_tess_no_pdfa
+def test_ghostscript_pdfa_failure(resources, outpdf):
+    p, _out, _err = run_ocrmypdf(
+        resources / 'francais.pdf',
+        outpdf,
+        '--plugin',
+        'tests/plugins/tesseract_noop.py',
+        '--plugin',
+        'tests/plugins/gs_pdfa_failure.py',
     )
     assert (
         p.returncode == ExitCode.pdfa_conversion_failed
     ), "Unexpected return when PDF/A fails"
 
 
-def test_ghostscript_feature_elision(spoof_no_tess_pdfa_warning, resources, outpdf):
-    check_ocrmypdf(resources / 'francais.pdf', outpdf, env=spoof_no_tess_pdfa_warning)
+def test_ghostscript_feature_elision(resources, outpdf):
+    check_ocrmypdf(
+        resources / 'francais.pdf',
+        outpdf,
+        '--plugin',
+        'tests/plugins/tesseract_noop.py',
+        '--plugin',
+        'tests/plugins/gs_feature_elision.py',
+    )
