@@ -17,7 +17,6 @@
 
 import pickle
 from math import isclose
-from tempfile import NamedTemporaryFile
 
 import img2pdf
 import pikepdf
@@ -84,8 +83,8 @@ def test_single_page_image(outdir):
     assert pdfimage.color == Colorspace.gray
 
     # DPI in a 1"x1" is the image width
-    assert isclose(pdfimage.xres, 8)
-    assert isclose(pdfimage.yres, 8)
+    assert isclose(pdfimage.dpi.x, 8)
+    assert isclose(pdfimage.dpi.y, 8)
 
 
 def test_single_page_inline_image(outdir):
@@ -104,19 +103,19 @@ def test_single_page_inline_image(outdir):
     info = pdfinfo.PdfInfo(filename)
     print(info)
     pdfimage = info[0].images[0]
-    assert isclose(pdfimage.xres, 8)
+    assert isclose(pdfimage.dpi.x, 8)
     assert pdfimage.color == Colorspace.gray
     assert pdfimage.width == 8
 
 
-def test_jpeg(resources, outdir):
+def test_jpeg(resources):
     filename = resources / 'c02-22.pdf'
 
     pdf = pdfinfo.PdfInfo(filename)
 
     pdfimage = pdf[0].images[0]
     assert pdfimage.enc == Encoding.jpeg
-    assert isclose(pdfimage.xres, 150)
+    assert isclose(pdfimage.dpi.x, 150)
 
 
 def test_form_xobject(resources):
@@ -132,13 +131,13 @@ def test_no_contents(resources):
 
     pdf = pdfinfo.PdfInfo(filename)
     assert len(pdf[0].images) == 0
-    assert pdf[0].has_text == False
+    assert not pdf[0].has_text
 
 
 def test_oversized_page(resources):
     pdf = pdfinfo.PdfInfo(resources / 'poster.pdf')
     image = pdf[0].images[0]
-    assert image.width * image.xres > 200, "this is supposed to be oversized"
+    assert image.width * image.dpi.x > 200, "this is supposed to be oversized"
 
 
 def test_pickle(resources):
@@ -150,22 +149,6 @@ def test_pickle(resources):
     pickle.dumps(pdf)
 
 
-def test_regex():
-    rx = pdfinfo.ghosttext.regex_remove_char_tags
-
-    must_match = [
-        b'<char bbox="0 108 0 108" c="/"/>',
-        b'<char bbox="0 108 0 108" c=">"/>',
-        b'<char bbox="0 108 0 108" c="X"/>',
-    ]
-    must_not_match = [b'<span stuff="c">', b'<span>', b'</span>', b'</page>']
-
-    for s in must_match:
-        assert rx.match(s)
-    for s in must_not_match:
-        assert not rx.match(s)
-
-
 def test_vector(resources):
     filename = resources / 'vector.pdf'
     pdf = pdfinfo.PdfInfo(filename)
@@ -185,11 +168,7 @@ def test_ocr_detection(resources):
 )
 def test_corrupt_font_detection(resources, testfile):
     filename = resources / testfile
-    with pytest.raises(NotImplementedError):
-        pdf = pdfinfo.PdfInfo(filename)
-        pdf[0].has_corrupt_text
-
-    pdf = pdfinfo.PdfInfo(filename, detailed_page_analysis=True)
+    pdf = pdfinfo.PdfInfo(filename, detailed_analysis=True)
     assert pdf[0].has_corrupt_text
 
 
